from numba import jit, prange
import numpy as np


@jit
def Lambda(x) -> np.ndarray:
    return 1 if x else 1/np.sqrt(2)


@jit
def precompute_cos() -> np.ndarray:
    cos_table = np.zeros((8, 8))
    for u in prange(8):
        for i in prange(8):
            cos_table[u, i] = np.cos(np.pi/8*(i+.5)*u)
    return cos_table

@jit
def forward_dct(X: np.ndarray):
    # precompute cosines
    N, M, _, _ = X.shape
    cos_table = precompute_cos()

    # iterate blocks
    Y = np.zeros((N, M, 8, 8))
    for h in prange(N):
        for w in prange(M):

            # iterate coefficients
            for u in prange(8):
                for v in prange(8):

                    # iterate pixels
                    Y[h, w, u, v] = 0
                    for i in prange(8):
                        for j in prange(8):

                            # compute Y_uv
                            Y[h, w, u, v] += (
                                X[h, w, i, j] *
                                cos_table[u, i] *
                                cos_table[v, j])
                    Y[h, w, u, v] *= (
                        np.sqrt(2/8) *
                        np.sqrt(2/8) *
                        Lambda(u) *
                        Lambda(v))
    return Y


@jit
def backward_dct(Y:np.ndarray) -> np.ndarray:
    """Backward DCT on image coefficients.

    :param X: input DCT coefficients
    :type X: np.ndarray
    :return: pixels
    :rtype: np.ndarray

    :Examples:

    You can use this together with the other jpeglib structures.

    >>> import jpeglib
    >>> jpeg = jpeglib.read_dct("image.jpeg")
    >>> x_Y = jpeglib.dct.backward(jpeg.Y)

    """
    # precompute cosines
    N, M, _, _ = Y.shape
    cos_table = precompute_cos()

    # iterate blocks
    X = np.zeros((N, M, 8, 8))
    for h in prange(N):
        for w in prange(M):

            # iterate pixels
            for i in prange(8):
                for j in prange(8):

                    # iterate coefficients
                    X[h, w, i, j] = 0
                    for u in prange(8):
                        for v in prange(8):

                            # compute X_ij
                            X[h, w, i, j] += (
                                Lambda(u) *
                                Lambda(v) *
                                Y[h, w, u, v] *
                                cos_table[u, i] *
                                cos_table[v, j])
                    X[h, w, i, j] *= (
                        np.sqrt(2/8) *
                        np.sqrt(2/8))
    return X




from __future__ import annotations
from dataclasses import dataclass
from typing import Union, Type


@dataclass
class CStruct:
    name: str
    """item name"""

    def __post_init__(self):
        self.index  # check if valid

    @classmethod
    def name_to_index(cls, name):
        raise NotImplementedError

    @classmethod
    def index_to_name(cls, index):
        raise NotImplementedError()

    @property
    def index(self) -> int:
        """Index getter."""
        return self.name_to_index(self.name)

    @property
    def name(self) -> str:
        """Name getter."""
        return self._name

    @name.setter
    def name(self, name: str):
        """Name setter."""
        self._name = name

    def __repr__(self) -> str:
        return '<CStruct %s>' % self.name

    def __str__(self) -> str:
        """Converts the class to str, returns name."""
        return self.name

    def __int__(self) -> int:
        """Converts the class to int, returns index."""
        return self.index

    @classmethod
    def from_index(cls, index):
        """Constructs the class from index.

        Uses the reverse mapping onto name.
        """
        name = cls.index_to_name(index)
        return cls(name=name)

    @classmethod
    def parse_input(cls, x: Union[str, CStruct]) -> int:
        # not specified
        if x is None:
            return -1
        # check string
        if isinstance(x, str):
            return cls(x).index
        # CStruct
        else:
            return x.index




import numpy as np

from ._common import Lambda


def _dct_ii(X: np.ndarray, u: int) -> float:
    """DCT coefficient at index u.

    :param X: input pixel data tensor
    :type X: np.ndarray
    :param u: index of DCT coefficient
    :type u: int
    :return: output DCT coefficient value
    :rtype: float

    :Examples:

    >>> _dct_ii([1,2,1], 1) # -> 2.72e-16
    """
    N = len(X)
    Yu = (
        np.sqrt(2/N) *
        Lambda(u) *
        sum(
            X[i] *
            np.cos(np.pi/N*(i+.5)*u)
            for i in range(N)
        )
    )
    return Yu


def _dct_iii(Y: np.ndarray, i: int) -> float:
    """Pixel value at index i.

    :param x: input DCT coefficient tensor
    :type x: np.ndarray
    :param i: index of the pixel
    :type i: int
    :return: output pixel value
    :rtype: float

    :Examples:

    >>> _dct_iii([2,1,-1], 1) # -> 1.97
    """
    N = len(Y)
    Xi = (
        np.sqrt(2/N) *
        sum(
            Lambda(u) *
            Y[u] *
            np.cos(np.pi/N*(i+.5)*u)
            for u in range(N)
        )
    )
    return Xi


def DCT1D(X: np.ndarray) -> np.ndarray:
    """Python implementation of 1D DCT transformation.

    :param X: pixel data in spatial domain
    :type X: np.ndarray
    :return: tensor of DCT coefficients
    :rtype: np.ndarray

    :Examples:

    >>> X = np.array([1,2,1])
    >>> Y = jpeglib.dct.DCT1D(X)
    """
    return [_dct_ii(X, k) for k in range(len(X))]


def iDCT1D(Y: np.ndarray):
    """Python implementation of 1D inverse DCT transformation.

    :param Y: tensor of DCT coefficients
    :type Y: np.ndarray
    :return: pixel data in spatial domain
    :rtype: np.ndarray

    :Examples:

    >>> X = np.array([1,2,1])
    >>> Y = jpeglib.dct.DCT1D(X)
    >>> jpeglib.dct.iDCT1D(Y) # -> reconstruction of X
    """
    return [_dct_iii(Y, k) for k in range(len(Y))]


__all__ = ["DCT1D", "iDCT1D"]




import ctypes
import dataclasses
import numpy as np
import os
import tempfile
from typing import List
import warnings

from ._bind import CJpegLib
from . import _jpeg


@dataclasses.dataclass
class DCTJPEG(_jpeg.JPEG):
    """JPEG instance to work with DCT domain."""

    Y: np.ndarray
    """luminance tensor"""
    Cb: np.ndarray
    """chrominance blue-difference tensor"""
    Cr: np.ndarray
    """chrominance red-difference tensor"""
    K: np.ndarray
    """black (cmyK) tensor"""
    qt: np.ndarray
    """quantization tensor"""
    quant_tbl_no: np.ndarray
    """assignment of quantization tables to components,
    (0 Y, 1 Cb, 1Cr) by default"""

    def _alloc_dct_component(self, i: int):
        return (
            ((ctypes.c_short * 64) *
             self.width_in_blocks(i)) *
            self.height_in_blocks(i)
        )()

    def load(self) -> tuple:
        """Function to allocate the buffer and read the image data.

        :return:
        luminance Y and chrominance Cb, Cr tensors
        and quantization table qt as Y,(Cb,Cr),qt
        :rtype: tuple

        :Example:

        >>> jpeg = jpeglib.read_dct("input.jpeg")
        >>> Y,(Cb,Cr),qt = jpeg.load()

        Function is internally called when

        >>> jpeg = jpeglib.read_dct("input.jpeg")
        >>> jpeg.Y # jpeg.load() called internally
        """
        # no content
        if self.content is None:
            return None, (None, None), None

        # allocate DCT components
        Y = self._alloc_dct_component(0)
        Cb, Cr = None, None
        if self.has_chrominance:  # has chrominance
            Cb = self._alloc_dct_component(1)
            Cr = self._alloc_dct_component(2)
        K = None
        if self.has_black:  # has black
            K = self._alloc_dct_component(3)
        qt = ((ctypes.c_short * 64) * 4)()
        _quant_tbl_no = (ctypes.c_ubyte * 4)()

        # write content into temporary file
        tmp = tempfile.NamedTemporaryFile(suffix='.jpeg', delete=False)
        tmp.write(self.content)
        tmp.flush()
        tmp.close()

        # call
        CJpegLib.read_jpeg_dct(
            path=str(self.path),
            srcfile=tmp.name,
            Y=Y,
            Cb=Cb,
            Cr=Cr,
            K=K,
            qt=qt,
            quant_tbl_no=_quant_tbl_no,
        )
        # clean up temporary file
        os.remove(tmp.name)

        # process
        def process_component(comp):
            comp_np = np.ctypeslib.as_array(comp)
            return comp_np.reshape((*comp_np.shape[:-1], 8, 8))
        qt = process_component(qt)
        self.Y = process_component(Y)
        if self.has_chrominance:
            self.Cb = process_component(Cb)
            self.Cr = process_component(Cr)
        if self.has_black:
            self.K = process_component(K)
        self.quant_tbl_no = np.array([
            _quant_tbl_no[i]
            for i in range(self.num_components)
        ])
        # crop
        self.qt = qt[:self.num_components]
        # return
        return self.Y, (self.Cb, self.Cr), self.qt

    def read_dct(self) -> tuple:
        warnings.warn('read_dct() is obsolete, use load()')
        return self.load()

    def write_dct(self, path: str = None, quality: int = -1):
        """Function to write DCT coefficients to JPEG file.

        Does not perform JPEG compression, writing DCT is lossless.

        :param path: Destination file name. If not given, source file is overwritten.
        :type path: str, optional
        :param quality: Compression quality, between 0 and 100. Special value -1 stands for using qt inside the instance or keeping libjpeg default.
        :type quality: int, optional

        :Example:

        >>> jpeg = jpeglib.read_spatial("input.jpeg")
        >>> jpeg.write_spatial("output.jpeg", quality=92)
        """  # noqa: E501
        # path
        dstfile = path if path is not None else self.path
        if dstfile is None:
            raise IOError('no destination file specified')

        # convert dct
        def process_component(comp):
            if comp is None:
                return None
            comp = comp.reshape((*comp.shape[:-2], 64))
            return np.ctypeslib.as_ctypes(comp.astype(np.int16))
        qt = process_component(self.qt)
        quant_tbl_no = np.array([-1]*4)
        if self.quant_tbl_no is not None:
            for i, q in enumerate(self.quant_tbl_no):
                quant_tbl_no[i] = q
            quant_tbl_no = np.ctypeslib.as_ctypes(quant_tbl_no.astype(np.int16))
        else:
            quant_tbl_no = None
        Y = process_component(self.Y)
        Cb = process_component(self.Cb)
        Cr = process_component(self.Cr)
        K = process_component(self.K)
        # quality and quantization table
        assert quality in set(range(-1, 101))
        if quality != -1:
            qt = None

        tmp = tempfile.NamedTemporaryFile(suffix='.jpeg', delete=False)
        if self.content is not None:
            # write content into temporary file
            tmp.write(self.content)
            tmp.flush()
        tmp.close()

        # call
        CJpegLib.write_jpeg_dct(
            srcfile=tmp.name if self.content is not None else None,
            dstfile=str(dstfile),
            Y=Y,
            Cb=Cb,
            Cr=Cr,
            K=K,
            image_dims=self.c_image_dims(),
            block_dims=self.c_block_dims(),
            in_color_space=self.jpeg_color_space.index,
            in_components=self.num_components,
            qt=qt,
            quality=quality,
            quant_tbl_no=quant_tbl_no,
            num_markers=self.num_markers,
            marker_types=self.c_marker_types(),
            marker_lengths=self.c_marker_lengths(),
            markers=self.c_markers(),
        )
        # clean up temporary file
        os.remove(tmp.name)

    @property
    def Y(self) -> np.ndarray:
        """Luminance tensor getter."""
        if self._Y is None:
            self.load()
        return self._Y

    @Y.setter
    def Y(self, Y: np.ndarray):
        """Luminance tensor setter."""
        self._Y = Y

    @property
    def Cb(self) -> np.ndarray:
        """Chrominance blue-difference tensor getter."""
        if self.has_chrominance and self._Cb is None:
            self.load()
        return self._Cb

    @Cb.setter
    def Cb(self, Cb: np.ndarray):
        """Chrominance blue-difference tensor setter."""
        self._Cb = Cb

    @property
    def Cr(self) -> np.ndarray:
        """Chrominance red-difference tensor getter."""
        if self.has_chrominance and self._Cr is None:
            self.load()
        return self._Cr

    @Cr.setter
    def Cr(self, Cr: np.ndarray):
        """Chrominance red-difference tensor setter."""
        self._Cr = Cr

    @property
    def K(self) -> np.ndarray:
        """Black tensor getter."""
        if self.has_black and self._K is None:
            self.load()
        return self._K

    @K.setter
    def K(self, K: np.ndarray):
        """Black tensor setter."""
        self._K = K

    @property
    def qt(self) -> np.ndarray:
        """Quantization table getter."""
        if self._qt is None:
            self.load()
        return self._qt

    @qt.setter
    def qt(self, qt: np.ndarray):
        """Quantization table setter."""
        self._qt = qt

    @property
    def quant_tbl_no(self) -> list:
        """Getter of assignment of quantization tables to components"""
        if self._quant_tbl_no is None:
            self.load()
        return self._quant_tbl_no

    @quant_tbl_no.setter
    def quant_tbl_no(self, quant_tbl_no: list):
        """Setter of assignment of quantization tables to components"""
        self._quant_tbl_no = quant_tbl_no

    def get_component_qt(self, idx: int) -> np.ndarray:
        """Getter of the quantization table of a component, dependent on assignment.

        :param idx: Component index.
        :type idx: int
        :return: quantization table of the component
        :rtype: np.ndarray

        :Examples:

        To retreive a component quantization table, simply type

        >>> jpeg = jpeglib.read_dct("input.jpeg")
        >>> # retrieval of component QTs
        >>> qtY = jpeg.get_component_qt(0)
        >>> qtCb = jpeg.get_component_qt(1)
        >>> qtCr = jpeg.get_component_qt(2)

        Usual assignment is Y having qt[0] and Cb and Cr sharing qt[1].
        However it may differ.

        >>> # retrieval of QTs at index
        >>> qt0 = jpeg.qt[0] # usually qtY
        >>> qt1 = jpeg.qt[1] # usually qtCb, qtCr (shared)

        E.g. ffmpeg uses a single quantization matrix for all three channels.
        """
        return self.qt[self.quant_tbl_no[idx]]

    def free(self):
        """Free the allocated tensors."""
        del self._Y
        del self._Cb
        del self._Cr
        del self._K
        del self._qt


@dataclasses.dataclass
class DCTJPEGio(DCTJPEG):
    """Class for compatiblity with jpegio."""

    coef_arrays: List
    """DCT coefficient arrays in jpegio format"""
    quant_tables: List
    """quantization tables in jpegio format"""

    def __post_init__(self):
        self._jpeg_to_jpegio()

    def _jpeg_to_jpegio(self):
        # colored image
        if self.has_chrominance:
            self._coef_arrays = [
                self._convert_dct_jpegio(self.Y),
                self._convert_dct_jpegio(self.Cb),
                self._convert_dct_jpegio(self.Cr)
            ]
            self._quant_tables = [
                self.get_component_qt(0).astype(np.int32),
                self.get_component_qt(1).astype(np.int32)
            ]
        # grayscale
        else:
            self._coef_arrays = [self._convert_dct_jpegio(self.Y)]
            self._quant_tables = [self.get_component_qt(0).astype(np.int32)]

    def _jpegio_to_jpeg(self):
        self.Y = self._convert_jpegio_dct(self.coef_arrays[0])
        if self.has_chrominance:
            self.Cb = self._convert_jpegio_dct(self.coef_arrays[1])
            self.Cr = self._convert_jpegio_dct(self.coef_arrays[2])
        if self.has_chrominance:
            self.qt = np.array([
                self.quant_tables[0],
                self.quant_tables[1]
            ], dtype=np.uint16)
            self.quant_tbl_no = np.array([0, 1, 1])
        else:
            self.qt = np.array([
                self.quant_tables[0]
            ], dtype=np.uint16)
            self.quant_tbl_no = np.array([0])

    def _convert_dct_jpegio(self, dct: np.ndarray) -> np.ndarray:
        return (
            dct
            .transpose((0, 3, 1, 2))
            .reshape((dct.shape[0]*dct.shape[2], dct.shape[1]*dct.shape[3]))
        ).astype(np.int32)

    def _convert_jpegio_dct(self, dct: np.ndarray) -> np.ndarray:
        return (
            dct
            .reshape((-1, 8, dct.shape[1]//8, 8))
            .transpose((0, 2, 3, 1))
        ).astype(np.int16)

    @property
    def coef_arrays(self) -> list:
        """Convertor of DCT coefficients to jpegio format."""
        return self._coef_arrays

    @coef_arrays.setter
    def coef_arrays(self, coef_arrays: list):
        """Setter of coefficient arrays in jpegio format."""
        self._coef_arrays = coef_arrays

    @property
    def quant_tables(self) -> list:
        """Convertor of quantization tables to jpegio format."""
        return self._quant_tables

    @quant_tables.setter
    def quant_tables(self, quant_tables: list):
        """Setter of quantization tables in jpegio format."""
        self._quant_tables = quant_tables

    def write(self, fpath: str, flags: int = -1, quality: int = -1):
        """Function to write DCT coefficients in jpegio format to JPEG file.

        Does not perform JPEG compression, writing DCT is lossless.

        :param fpath: Destination file name. If not given, source file is overwritten.
        :type fpath: str, optional
        :param flags: Flags. For backward compatibility, does not make a difference.
        :type flags: int
        :param quality: Compression quality, between 0 and 100. Special value -1 stands for using qt inside the instance or keeping libjpeg default.
        :type quality: int, optional

        :Example:

        >>> jpeg = jpeglib.read_dct("input.jpeg")
        >>> jpeg = jpeglib.to_jpegio(jpeg)
        >>> jpeg.write("output.jpeg", quality=92)
        """  # noqa: E501
        # pass data from jpegio to jpeg
        self._jpegio_to_jpeg()
        # write
        self.write_dct(path=fpath, quality=quality)


def to_jpegio(jpeg: DCTJPEG) -> DCTJPEGio:
    """Convertor of object of :class:`DCTJPEG` to :class:`DCTJPEGio`.

    When :class:`DCTJPEG` is converted to :class:`DCTJPEGio`,
    the data objects :class:`DCTJPEG` are not updated until writing.
    This behavior will be changed in the future.

    :param jpeg: JPEG object to convert
    :type jpeg: :class:`DCTJPEG`
    :return: converted JPEG object
    :rtype: :class:`DCTJPEGio`

    :Example:

    >>> jpeg = jpeglib.read_dct("input.jpeg")
    >>> jpeg = jpeglib.to_jpegio(jpeg)
    >>> jpeg.coef_arrays; jpeg.quant_tables
    """
    vals = {field.name: getattr(jpeg, field.name) for field in dataclasses.fields(jpeg)}
    return DCTJPEGio(**vals)




import numpy as np

from ._common import Lambda


def _dct_ii(X: np.ndarray, u: int, v: int):
    """DCT coefficient at index u,v.

    :param X: input pixel data tensor
    :type X: np.ndarray
    :param u: horizontal index of DCT coefficient
    :type u: int
    :param v: vertical index of DCT coefficient
    :type v: int
    :return: output DCT coefficient value
    :rtype: float

    :Examples:

    >>> X = np.array([[1,2],[3,1]])
    >>> _dct_ii(X,1,0) # -> 2.72e-16
    """
    N, M = X.shape
    Yuv = (
        np.sqrt(2/N) *
        np.sqrt(2/M) *
        Lambda(u) *
        Lambda(v) *
        sum(
            X[i, j] *
            np.cos(np.pi/N*(i+.5)*u) *
            np.cos(np.pi/N*(j+.5)*v)
            for i in range(N) for j in range(M)
        )
    )
    return Yuv


def _dct_iii(Y, i, j):
    """Pixel value at index i,j.

    :param x: input DCT coefficient tensor
    :type x: np.ndarray
    :param i: horizontal index of the pixel
    :type i: int
    :param j: vertical index of the pixel
    :type j: int
    :return: output value
    :rtype: float
    """
    N, M = Y.shape
    Xij = (
        np.sqrt(2/N) *
        np.sqrt(2/M) *
        sum(
            (
                Lambda(u) *
                Lambda(v) *
                Y[u, v] *
                np.cos(np.pi/N*(i + .5)*u) *
                np.cos(np.pi/N*(j + .5)*v)
            )
            for u in range(N) for v in range(M)
        )
    )
    return Xij


def DCT2D(X: np.ndarray):
    """Python implementation of 2D DCT transformation.

    :param X: pixel data in spatial domain
    :type X: np.ndarray
    :return: tensor of DCT coefficients
    :rtype: np.ndarray

    :Examples:

    >>> X = np.array([[1,2],[3,1]])
    >>> Y = jpeglib.dct.DCT2D(X)
    """
    return np.array([
        [
            _dct_ii(X, u, v)
            for v in range(X.shape[1])
        ]
        for u in range(X.shape[0])
    ])


def iDCT2D(Y: np.ndarray):
    """Python implementation of 2D inverse DCT transformation.

    :param Y: tensor of DCT coefficients
    :type Y: np.ndarray
    :return: pixel data in spatial domain
    :rtype: np.ndarray

    :Examples:

    >>> X = np.array([[1,2],[3,1]])
    >>> Y = jpeglib.dct.DCT2D(X)
    >>> jpeglib.dct.iDCT2d(Y) # -> reconstruction of X
    """
    return np.array([
        [
            _dct_iii(Y, i, j)
            for j in range(Y.shape[1])
        ]
        for i in range(Y.shape[0])
    ])


__all__ = ["DCT2D", "iDCT2D"]




import ctypes
from dataclasses import dataclass
import numpy as np
import os
import tempfile
from typing import Union
import warnings

from ._bind import CJpegLib
from ._jpeg import JPEG
from ._colorspace import Colorspace
from ._dithermode import Dithermode
from ._dctmethod import DCTMethod


@dataclass
class SpatialJPEG(JPEG):
    """JPEG instance to work in spatial domain."""
    spatial: np.ndarray
    """pixel data tensor"""
    color_space: Colorspace
    """color space of the pixel data"""
    # needed for compression, not actual props
    # dither_mode: Dithermode
    # dct_method: DCTMethod
    flags: list

    def _alloc_spatial(self, channels: int = None):
        if channels is None:
            channels = self.color_space.channels
        return (((ctypes.c_ubyte * self.width) * self.height) * channels)()

    def load(
        self,
        dct_method: Union[DCTMethod, str] = None,
        dither_mode: Union[Dithermode, str] = None,
    ) -> np.ndarray:

        # colorspace
        if self.color_space is None:
            if self.jpeg_color_space == Colorspace('JCS_CMYK'):
                self.color_space = Colorspace('JCS_CMYK')
            else:
                self.color_space = Colorspace('JCS_RGB')
            # self.color_space = self.jpeg_color_space
        # dither mode
        dither_mode = Dithermode.parse_input(dither_mode)
        # dct method
        dct_method = DCTMethod.parse_input(dct_method)

        # allocate spatial
        spatial = self._alloc_spatial(self.color_space.channels)

        # write content into temporary file
        tmp = tempfile.NamedTemporaryFile(suffix='.jpeg', delete=False)
        tmp.write(self.content)
        tmp.flush()
        tmp.close()

        # call
        CJpegLib.read_jpeg_spatial(
            path=str(self.path),
            srcfile=tmp.name,
            spatial=spatial,
            colormap=self.jpeg_color_space.index,
            in_colormap=None,  # support of color quantization
            out_color_space=self.color_space.index,
            dither_mode=dither_mode,
            dct_method=dct_method,
            flags=self.flags,
        )
        # clean up temporary file
        os.remove(tmp.name)
        # process
        self.spatial = (
            np.ctypeslib.as_array(spatial)
            .astype(np.ubyte)
            .reshape(self.height, -1, self.color_space.channels)
        )

        return self.spatial

    def read_spatial(self) -> np.ndarray:
        warnings.warn('read_spatial() is obsolete, use load()')
        return self.load()

    def write_spatial(
        self,
        path: str = None,
        qt: Union[int, np.ndarray] = None,
        quant_tbl_no: np.ndarray = None,
        base_quant_tbl_idx: int = None,
        dct_method: Union[str, DCTMethod] = None,
        # dither_mode: Dithermode = None,
        smoothing_factor: int = None,
        flags: list = []
    ):
        """Writes a spatial image representation (i.e. RGB) to a file.

        :param path: Destination file name. If not given, source file is overwritten.
        :type path: str, optional
        :param qt: Compression quality, can be integer 0-100 or a tensor with quantization tables. Defaultly -1 (default factor kept).
        :type qt: int | numpy.ndarray, optional
        :param quant_tbl_no: assignment of quantization tables to components, (0 Y, 1 Cb, 1Cr) by default
        :type quant_tbl_no: numpy.ndarray, optional
        :param base_quant_tbl_idx: base QT to scale, only supported in MozJPEG 3+
        :type base_quant_tbl_idx: int, optional
        :param dct_method: DCT method, must be accepted by :class:`_dctmethod.DCTMethod`. If not given, using the libjpeg default.
        :type dct_method: str | :class:`_dctmethod.DCTMethod`, optional
        :param smoothing_factor: Smoothing factor, between 0 and 100. Using default from libjpeg by default.
        :type smoothing_factor: int, optional
        :param flags: Bool compression parameters as str. If not given, using the libjpeg default. More at `glossary <https://jpeglib.readthedocs.io/en/latest/glossary.html#flags>`_.
        :type flags: list, optional

        :Example:

        >>> jpeg = jpeglib.read_spatial("input.jpeg")
        >>> jpeg.write_spatial("output.jpeg", qt=75)

        Grayscale JPEG (only lumo channel) using libjpeg color conversion

        >>> x = np.random.randint(0,255,(16,16,3),dtype=np.uint8)
        >>> im = jpeglib.from_spatial(x)
        >>> im.jpeg_color_space = jpeglib.Colorspace('JCS_GRAYSCALE')
        >>> im.write_spatial("hello.jpeg")
        """  # noqa: E501
        # colorspace
        if self.jpeg_color_space is not None:
            jpeg_color_space = self.jpeg_color_space
        else:
            jpeg_color_space = self.color_space
        num_components = (ctypes.c_int*2)(
            self.color_space.channels, jpeg_color_space.channels)
        jpeg_color_space = (ctypes.c_int*2)(
            self.color_space.index, jpeg_color_space.index)
        # path
        dstfile = path if path is not None else self.path
        if dstfile is None:
            raise IOError('no destination file specified')
        # dct method
        dct_method = DCTMethod.parse_input(dct_method)
        # quality
        # use default of library
        if qt is None:
            quality, qt = -1, None
        else:
            # quality factor
            try:
                quality, qt = int(qt), None
            # quantization table
            except TypeError:
                quality, qt = -1, np.ctypeslib.as_ctypes(qt.astype(np.uint16))
                if quant_tbl_no is not None:
                    quant_tbl_no = np.ctypeslib.as_ctypes(np.array(quant_tbl_no).astype(np.int16))

        # process
        spatial = np.ctypeslib.as_ctypes(
            self.spatial.reshape(
                self.color_space.channels,
                self.height,
                self.width,
            )
        )
        # call
        CJpegLib.write_jpeg_spatial(
            dstfile=str(dstfile),
            spatial=spatial,
            image_dims=self.c_image_dims(),
            jpeg_color_space=jpeg_color_space,
            num_components=num_components,
            dct_method=dct_method,
            samp_factor=self.c_samp_factor(),
            qt=qt,
            quality=quality,
            quant_tbl_no=quant_tbl_no,
            base_quant_tbl_idx=base_quant_tbl_idx,
            smoothing_factor=smoothing_factor,
            num_markers=self.num_markers,
            marker_types=self.c_marker_types(),
            marker_lengths=self.c_marker_lengths(),
            markers=self.c_markers(),
            flags=flags,
        )

    @property
    def spatial(self) -> np.ndarray:
        if self._spatial is None:
            self.load()
        return self._spatial

    @spatial.setter
    def spatial(self, spatial: np.ndarray):
        self._spatial = spatial

    @property
    def color_space(self) -> np.ndarray:
        return self._color_space

    @color_space.setter
    def color_space(self, color_space: Colorspace):
        self._color_space = color_space

    @property
    def channels(self) -> int:
        if self._color_space is None:
            return None
        return self._color_space.channels

    @property
    def dither_mode(self) -> np.ndarray:
        return self._dither_mode

    @dither_mode.setter
    def dither_mode(self, dither_mode: Dithermode):
        self._dither_mode = dither_mode

    @property
    def dct_method(self) -> np.ndarray:
        return self._dct_method

    @dct_method.setter
    def dct_method(self, dct_method: DCTMethod):
        self._dct_method = dct_method

    @property
    def flags(self) -> list:
        return self._flags

    @flags.setter
    def flags(self, flags: list):
        self._flags = flags

    def free(self):
        """Free the allocated tensors."""
        del self._spatial



"""
Python envelope for the popular C library libjpeg for handling JPEG files.

It offers full control over compression and decompression
and exposes DCT coefficients and quantization tables.
"""

# functions
from .functional import read_dct, read_spatial, from_spatial, from_dct

# jpeg objects
from .dct_jpeg import DCTJPEG, DCTJPEGio, to_jpegio
from .spatial_jpeg import SpatialJPEG

# cstructs
from ._colorspace import Colorspace
from ._dctmethod import DCTMethod
from ._dithermode import Dithermode
from ._marker import Marker

# libjpeg versions
from .version import version

# DCT implementation
from . import ops

# for unit tests
from ._timer import Timer

# package version
import pkg_resources
try:
    __version__ = pkg_resources.get_distribution("jpeglib").version
except pkg_resources.DistributionNotFound:
    __version__ = None

# set default version
try:
    version.set('6b')
except IndexError:
    import logging
    logging.warning('found versions: ' + ' '.join(version.versions()))
    raise RuntimeError('invalid installation, version 6b not found')

__all__ = [
    'read_dct', 'read_spatial', 'from_spatial', 'from_dct', 'to_jpegio',
    'SpatialJPEG', 'DCTJPEG', 'DCTJPEGio',
    'Colorspace', 'DCTMethod', 'Dithermode', 'Marker',
    'version','ops',
    'Timer', 'dct',
    '__version__',
]




import ctypes
import os
import pathlib
import re
from typing import List

from . import cjpeglib


class CJpegLib:

    @classmethod
    def jpeg_lib_version(cls):
        return cls.get().jpeg_lib_version()

    @classmethod
    def read_jpeg_info(
        cls,
        srcfile: str,
        block_dims,
        image_dims,
        num_components,
        samp_factor,
        jpeg_color_space: int,
        marker_lengths,
        marker_types,
        flags: List[str],
    ):
        status = cls.get().read_jpeg_info(
            cls.cstr(srcfile),
            block_dims,
            image_dims,
            num_components,
            samp_factor,
            jpeg_color_space,
            marker_lengths,
            marker_types,
            flags
        )
        if status == 0:
            raise IOError(f"reading info of {srcfile} failed")

    @classmethod
    def read_jpeg_markers(cls, srcfile: str, markers):
        status = cls.get().read_jpeg_markers(cls.cstr(srcfile), markers)
        if status == 0:
            raise IOError(f"reading markers of {srcfile} failed")

    @classmethod
    def read_jpeg_dct(
        cls,
        srcfile: str,
        Y,
        Cb,
        Cr,
        K,
        qt,
        quant_tbl_no,
        path=None,
    ):
        if path is None:
            path = srcfile
        status = cls.get().read_jpeg_dct(
            cls.cstr(srcfile),
            Y,
            Cb,
            Cr,
            K,
            qt,
            quant_tbl_no
        )
        if status == 0:
            print(f"{path=} {srcfile=}")
            raise IOError(f"reading of {path} DCT failed")

    @classmethod
    def write_jpeg_dct(
        cls,
        srcfile: str,
        dstfile: str,
        Y,
        Cb,
        Cr,
        K,
        image_dims,
        block_dims,
        in_color_space,
        in_components,
        qt,
        quality,
        quant_tbl_no,
        num_markers,
        marker_types,
        marker_lengths,
        markers,
    ):
        status = cls.get().write_jpeg_dct(
            cls.cstr(srcfile),
            cls.cstr(dstfile),
            Y,
            Cb,
            Cr,
            K,
            image_dims,
            block_dims,
            in_color_space,
            in_components,
            qt,
            quality,
            quant_tbl_no,
            num_markers,
            marker_types,
            marker_lengths,
            markers
        )
        if status == 0:
            raise IOError(f"writing DCT to {dstfile} failed")

    @classmethod
    def print_jpeg_params(cls, srcfile: str):
        status = cls.get().print_jpeg_params(cls.cstr(srcfile))
        if status == 0:
            raise IOError(f"reading of {srcfile} failed")

    @classmethod
    def read_jpeg_spatial(
        cls,
        srcfile: str,
        spatial,
        colormap,
        in_colormap,
        out_color_space,
        dither_mode,
        dct_method,
        flags,
        path=None,
    ):
        if path is None:
            path = srcfile
        status = cls.get().read_jpeg_spatial(
            cls.cstr(srcfile),
            spatial,
            colormap,
            in_colormap,
            out_color_space,
            dither_mode,
            dct_method,
            cls.flags_to_mask(flags)
        )
        if status == 0:
            raise IOError(f"reading of {path} spatial failed")

    @classmethod
    def write_jpeg_spatial(
        cls,
        dstfile: str,
        spatial,
        image_dims,
        jpeg_color_space,
        num_components,
        dct_method,
        samp_factor,
        qt,
        quality,
        quant_tbl_no,
        base_quant_tbl_idx,
        smoothing_factor,
        num_markers: int,
        marker_types,
        marker_lengths,
        markers,
        flags: List[str],
    ):
        status = cls.get().write_jpeg_spatial(
            cls.cstr(dstfile),
            spatial,
            image_dims,
            jpeg_color_space,
            num_components,
            dct_method,
            samp_factor,
            qt,
            cls.factor(quality),
            quant_tbl_no,
            cls.factor(base_quant_tbl_idx),
            cls.factor(smoothing_factor),
            num_markers,
            marker_types,
            marker_lengths,
            markers,
            cls.flags_to_mask(flags)
        )
        if status == 0:
            raise IOError(f"writing spatial to {dstfile} failed")

    MASKS = {
        "DO_FANCY_SAMPLING": (0b1 << 0),
        "DO_FANCY_UPSAMPLING": (0b1 << 0),
        "DO_FANCY_DOWNSAMPLING": (0b1 << 0),
        "DO_BLOCK_SMOOTHING": (0b1 << 2),
        "TWO_PASS_QUANTIZE": (0b1 << 4),
        "ENABLE_1PASS_QUANT": (0b1 << 6),
        "ENABLE_EXTERNAL_QUANT": (0b1 << 8),
        "ENABLE_2PASS_QUANT": (0b1 << 10),
        "OPTIMIZE_CODING": (0b1 << 12),
        "PROGRESSIVE_MODE": (0b1 << 14),
        "QUANTIZE_COLORS": (0b1 << 16),
        "ARITH_CODE": (0b1 << 18),
        "WRITE_JFIF_HEADER": (0b1 << 20),
        "WRITE_ADOBE_MARKER": (0b1 << 22),
        "CCIR601_SAMPLING": (0b1 << 24),
        "FORCE_BASELINE": (0b1 << 26),
    }

    @classmethod
    def flags_to_mask(cls, flags: List[str]):
        mask = 0xFFFFFFFF
        if flags is None:
            return mask
        for flag in flags:
            # parse sign
            sign = '-' if flag[0] == '-' else '+'
            if not flag[0].isalpha():
                flag = flag[1:]
            # get flags
            flagbit = cls.MASKS[flag.upper()]
            defbit = cls.MASKS[flag.upper()] << 1
            # map
            mask ^= defbit  # reset default value
            if sign == '-':
                mask ^= (flagbit)  # erase bit

        return ctypes.c_ulonglong(mask)

    @classmethod
    def mask_to_flags(cls, mask: int):
        flags = []
        bitmask = mask[0]
        # PROGRESSIVE_MODE = 0b00100
        # mask = ??1?? or ??0??
        if ((cls.MASKS["PROGRESSIVE_MODE"]) & bitmask) != 0:
            flags.append("PROGRESSIVE_MODE")

        return flags

    @classmethod
    def factor(cls, factor):
        if factor is None:
            factor = -1
        return ctypes.c_short(factor)

    _lib = None
    version = None

    @classmethod
    def get(cls):
        # connect to library
        if cls._lib is None:
            cls._lib = cls._bind_lib()
        # return library
        return cls._lib

    @classmethod
    def set_version(cls, version):
        cls._lib = cls._bind_lib(version=version)

    @classmethod
    def get_version(cls):
        return cls.version

    @classmethod
    def _versions(cls):
        so_files = [
            f
            for f in os.listdir(list(cjpeglib.__path__)[0])
            if re.fullmatch(r'cjpeglib_.*\.(.*\.so|pyd)', f)
        ]
        return so_files

    @classmethod
    def versions(cls):
        # list DLLs
        vs = [re.search(r'cjpeglib_[^.]*\.(.*\.so|pyd)', f)[0] for f in cls._versions()]
        # parse versions
        vs = [re.search(r'(?<=cjpeglib_)[^.]*', f)[0] for f in cls._versions()]
        return vs

    @classmethod
    def _bind_lib(cls, version='6b'):
        # path of the library
        so_files = [f for f in cls._versions() if re.fullmatch(
            f'cjpeglib_{version}' + r'\.(.*\.so|pyd)', f)]
        try:
            so_file = so_files[0]
        except IndexError:
            so_file = None
        if so_file is None:
            raise RuntimeError(f'version "{version}" not found')
        libname = str(pathlib.Path(list(cjpeglib.__path__)[0]) / so_file)
        # libname = str(os.path.join(list(cjpeglib.__path__)[0], so_file))
        # connect
        cjpeglib_dylib = ctypes.CDLL(libname)
        cls.version = version
        return cjpeglib_dylib

    @staticmethod
    def cstr(s):
        if s is None:
            return None
        return ctypes.c_char_p(s.encode('utf-8'))




import numpy as np


def blockify_8x8(x: np.ndarray) -> np.ndarray:
    """Splits the spatial domain into 8x8 tiles.

    The output has shape (N/8, M/8, 8, 8, channels).

    :param x: input pixel array
    :type x: np.ndarray
    :return: tiled pixels
    :rtype: np.ndarray

    :Examples:

    Example of usage with jpeglib

    >>> import jpeglib
    >>> im = jpeglib.read_spatial('image.jpeg')
    >>> r = jpeglib.ops.blockify_8x8(im.x)
    """
    # get shape
    N, M, _ = x.shape
    # split along each axis (separately)
    xh = np.split(x, (N+7)//8, axis=0)
    xhv = [np.split(r, (M+7)//8, axis=1) for r in xh]
    # merge
    xb = np.array(xhv)
    return xb

def deblockify_8x8(xb:np.ndarray) -> np.ndarray:
    """Merges split spatial domain (8x8 blocks).

    The output has shape (N, M) for input (N/8, M/8, 8, 8)

    :param xb: tiled pixels
    :type xb: np.ndarray
    :return: spatial domain representation
    :rtype: np.ndarray

    :Examples:

    Example of usage with jpeglib

    >>> import jpeglib
    >>> im = jpeglib.read_dct('image.jpeg')
    >>> X = jpeglib.backward_dct(im.Y)
    >>> r = jpeglib.ops.deblockify_8x8(X)
    """
    return (
        np.einsum('abcd->adbc', xb)
            .reshape(xb.shape[0]*8, xb.shape[1]*8)
    )

def luminance(x: np.ndarray) -> np.ndarray:
    """Converts RGB into luminance / grayscale.

    :param x: input rgb pixel array
    :type x: np.ndarray
    :return: grayscale pixel array
    :rtype: np.ndarray

    :Examples:

    Example of usage with jpeglib

    >>> import jpeglib
    >>> im = jpeglib.read_spatial('image.jpeg')
    >>> g = jpeglib.ops.luminance(im.x)
    """
    lumo_contributions = np.array([0.2989, 0.5870, 0.1140])
    return np.expand_dims(x @ lumo_contributions, 2)


def grayscale(x: np.ndarray) -> np.ndarray:
    return luminance(x=x)




from typing import List


def GCD(*args) -> float:
    """Computes greatest common divider (gcd) of the positional arguments."""
    # replace 0s with 1s
    args = [a if a != 0 else 1 for a in args]
    # all divisors
    divisors = [
        i
        for i in range(1, min(args)+1)
        if all([x % i == 0 for x in args])
    ]
    # choose the maximal
    gcd = max(divisors)
    return gcd


def Jab_to_factors(Jab) -> List:
    """Converts J:a:b notation to the 6 factors, used by libjpeg.

    Read more about the notations in
    `Tuomas Siipola's article <https://zpl.fi/chroma-subsampling-and-jpeg-sampling-factors/>`_
    or `glossary <https://jpeglib.readthedocs.io/en/latest/glossary.html#jpeg-sampling-factor>`_.

    :param Jab:
    List of three numbers, J:a:b notation of subsampling.
    :type dstfile: list
    :return:
    Subsampling factors, as list of three lists, each containing 2 numbers.
    :rtype: list

    :Example:

    >>> Jab_to_factors([4,2,0]) # -> [[2,2],[1,1],[1,1]]
    """  # noqa: E501
    # parse input
    J, a, b = Jab
    assert J in {1, 2, 3, 4}
    if J == 4:
        assert a in {4, 2, 1, 0}
    elif J == 3:
        assert a in {3, 1, 0}
    else:
        assert a in {J, 1, 0}
    assert b in {a, 0}
    # chroma dimensions
    Y = [J, 2]
    Cb = Cr = [a, (a == b) + 1]
    # normalize by GCD
    gcd0 = GCD(Y[0], Cb[0], Cr[0])
    gcd1 = GCD(Y[1], Cb[1], Cr[1])
    factors = [
        [int(Y[0] / gcd0), int(Y[1] / gcd1)],
        [int(Cb[0] / gcd0), int(Cb[1] / gcd1)],
        [int(Cr[0] / gcd0), int(Cr[1] / gcd1)],
    ]
    return factors


if __name__ == '__main__':
    Jab_to_factors([4, 4, 4])
    Jab_to_factors([4, 4, 0])
    Jab_to_factors([4, 2, 2])
    Jab_to_factors([4, 2, 0])
    Jab_to_factors([4, 1, 1])
    Jab_to_factors([4, 1, 0])
    Jab_to_factors([3, 3, 3])
    Jab_to_factors([3, 3, 0])
    Jab_to_factors([3, 1, 1])
    Jab_to_factors([3, 1, 0])




from dataclasses import dataclass
from ._cstruct import CStruct


@dataclass
class DCTMethod(CStruct):

    _J_DCT_METHOD = {
        "JDCT_ISLOW":   0,  #
        "JDCT_IFAST":   1,  #
        "JDCT_FLOAT":   2,  #
    }

    @classmethod
    def J_DCT_METHOD(cls) -> dict:
        return cls._J_DCT_METHOD

    @classmethod
    def iJ_DCT_METHOD(cls) -> dict:
        return {v: k for k, v in cls._J_DCT_METHOD.items()}

    @classmethod
    def name_to_index(cls, name: str) -> str:
        return cls.J_DCT_METHOD()[name]

    @classmethod
    def index_to_name(cls, index: int) -> str:
        return cls.iJ_DCT_METHOD()[index]

    def __repr__(self) -> str:
        return '<DCTMethod %s>' % self.name




import codecs
import ctypes
import os
from pathlib import Path
import re
import setuptools
import setuptools.command.build_ext
import sys

# wheel builder
try:
    from wheel.bdist_wheel import bdist_wheel

    class bdist_wheel_abi3(bdist_wheel):
        def get_tag(self):
            python, abi, plat = super().get_tag()
            if python.startswith("cp"):
                return "cp38", "abi3", plat
            return python, abi, plat

    custom_bdist_wheel = {'bdist_wheel': bdist_wheel_abi3}
except ModuleNotFoundError:
    custom_bdist_wheel = {}

# versions
__version__ = os.environ.get('VERSION_NEW', '0.11.4')
libjpeg_versions = {
    '6b': (None, 60),
    '7': (None, 70),
    '8': (None, 80),
    '8a': (None, 80),
    '8b': (None, 80),
    '8c': (None, 80),
    '8d': (None, 80),
    '9': (None, 90),
    '9a': (None, 90),
    '9b': (None, 90),
    '9c': (None, 90),
    '9d': (None, 90),
    '9e': (None, 90),
    'turbo120': ('1.2.0', 3120),
    'turbo130': ('1.3.0', 3130),
    'turbo140': ('1.4.0', 3140),
    'turbo150': ('1.5.0', 3150),
    'turbo200': ('2.0.0', 3200),
    'turbo210': ('2.1.0', 3210),
    'mozjpeg101': ('1.0.1', 6101),
    'mozjpeg201': ('2.0.1', 6201),
    'mozjpeg300': ('3.0.0', 6300),
    'mozjpeg403': ('4.0.3', 6403),
}

# requirements
try:
    with open('requirements.txt') as f:
        reqs = f.read().splitlines()
except FileNotFoundError:
    reqs = ['numpy']

# description
with codecs.open("README.md", "r", encoding="UTF-8") as fh:
    long_description = fh.read()

# create version dependent extensions
cfiles, hfiles = {}, {}
cjpeglib = {}
for v in libjpeg_versions:

    # library-dependent
    is_moz = v[:3] == "moz"
    is_turbo_moz = v[:5] == "turbo" or is_moz

    # name of C library
    clib = f'jpeglib/cjpeglib/{v}'

    # create missing
    package_name = 'libjpeg'
    (Path(clib) / 'jconfig.h').touch()
    (Path(clib) / 'config.h').touch()
    if True:  # not (Path(clib) / 'vjpeglib.h').exists():
        with open(Path(clib) / 'vjpeglib.h', 'w') as f:
            f.write('#include "jpeglib.h"')
    if is_turbo_moz:
        package_name += '-turbo'
        (Path(clib) / 'jconfigint.h').touch()
        if is_moz:
            package_name = 'mozjpeg'
            (Path(clib) / 'config.h').touch()

    # get all files
    files = [
        f'{clib}/{f}'
        for f in os.listdir(clib)
        if re.fullmatch(r'.*\.(c|h)', f)
    ]
    # exclude files
    for excluded_module in [
        # platform dependents
        'jmemdos', 'jmemmac', 'jmemansi',
        'ansi2knr', 'ckconfig', 'jmemname',
        # executables
        'djpeg', 'cjpeg', 'rdjpgcom',
        'wrjpgcom', 'cdjpeg', 'jpegtran',
        # others
        'rdbmp', 'wrbmp', 'rdcolmap',
        'rdppm', 'wrppm', 'rdtarga',
        'wrtarga', 'rdrle', 'wrrle',
        'rdgif', 'wrgif', 'rdswitch',
        # example
        'example',
        #
        'cjpegalt', 'djpegalt',
        # turbo
        'jccolext', 'jdcolext', 'jdcol565',
        'jstdhuff', 'jdmrg565', 'jdmrgext',
        'jcstest', 'tjunittest', 'tjbench',
        'turbojpeg-jni', 'turbojpeg', 'turbojpegl',
        'jpegut', 'jpgtest',
        # mozjpeg
        'bmp', 'jpegyuv',
    ]:
        lim = -2 - len(excluded_module)
        files = [f for f in files if f[lim:-2] != excluded_module]
    # split to sources and headers
    cfiles[v] = [f for f in files if f[-2:] == '.c']
    hfiles[v] = [f for f in files if f[-2:] == '.h']
    sources = cfiles[v]
    hfiles[v].append('jpeglib/cjpeglib/cjpeglib_common.h')
    hfiles[v].append('jpeglib/cjpeglib/cjpeglib_common_flags.h')
    hfiles[v].append('jpeglib/cjpeglib/cjpeglib_common_markers.h')
    sources.append('jpeglib/cjpeglib/cjpeglib_dct.cpp')
    sources.append('jpeglib/cjpeglib/cjpeglib_spatial.cpp')
    sources.append('jpeglib/cjpeglib/cjpeglib_common_flags.c')
    sources.append('jpeglib/cjpeglib/cjpeglib_common_markers.cpp')
    sources.append('jpeglib/cjpeglib/cjpeglib_common.cpp')

    # define macros
    macros = [
        ("BITS_IN_JSAMPLE", 8),
        ("HAVE_UNSIGNED_CHAR", 1),
        ("HAVE_STDLIB_H", 1),
        ("LIBVERSION", libjpeg_versions[v][1]),
        ("HAVE_PROTOTYPES", 1),
        ("Py_LIMITED_API", "0x03080000"),
    ]
    # turbo/moz-only macros
    if is_turbo_moz:
        macros += [
            ("INLINE", "__inline__" if not sys.platform.startswith("win") else "__inline"),
            ("PACKAGE_NAME", f"\"{package_name}\""),
            ("BUILD", "\"unknown\""),
            ("VERSION", f"\"{libjpeg_versions[v][0]}\""),
            ("SIZEOF_SIZE_T", int(ctypes.sizeof(ctypes.c_size_t))),
            ("THREAD_LOCAL", "__thread"),
            ("C_ARITH_CODING_SUPPORTED", 1),
            ("D_ARITH_CODING_SUPPORTED", 1),
            ("JPEG_LIB_VERSION", 70),  # 70), # turbo 2.1.0
        ]
        # moz-only macros
        if is_moz:
            macros += [
                # ("JPEG_LIB_VERSION", 69),
                ('MEM_SRCDST_SUPPORTED', 1),
            ]
        # if not is_moz or libjpeg_versions[v][1] <= 403:
        #     macros += [
        #         ("JPEG_LIB_VERSION", 70),  # 70), # turbo 2.1.0
        #     ]

    # define the extension
    cjpeglib[v] = setuptools.Extension(
        name=f"jpeglib/cjpeglib/cjpeglib_{v}",
        library_dirs=['./jpeglib/cjpeglib', f'./{clib}'],
        include_dirs=['./jpeglib/cjpeglib', f'./{clib}'],
        sources=sources,
        headers=hfiles[v],
        define_macros=macros,
        extra_compile_args=[] if sys.platform.startswith("win") else ["-fPIC", "-g"],
        # language='C++',
        py_limited_api=True,
    )


# extension builder
class custom_build_ext(setuptools.command.build_ext.build_ext):
    def get_export_symbols(self, ext):
        parts = ext.name.split(".")
        if parts[-1] == "__init__":
            initfunc_name = "PyInit_" + parts[-2]
        else:
            initfunc_name = "PyInit_" + parts[-1]

    def build_extensions(self):
        setuptools.command.build_ext.build_ext.build_extensions(self)
        setuptools.command.build_ext.build_ext.get_export_symbols = self.get_export_symbols

# define package
setuptools.setup(
    name='jpeglib',
    version=__version__,
    author=u'Martin Beneš',
    author_email='martinbenes1996@gmail.com',
    description="Python envelope for the popular C library" +
                "libjpeg for handling JPEG files.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    license='MPL',
    project_urls={
        "Homepage": "https://pypi.org/project/jpeglib/",
        "Documentation": 'https://jpeglib.readthedocs.io/en/latest/',
        "Source": "https://github.com/martinbenes1996/jpeglib/",
    },
    keywords=['jpeglib', 'jpeg', 'jpg', 'libjpeg', 'compression',
              'decompression', 'dct-coefficients', 'dct'],
    install_requires=reqs,
    package_dir={'': '.'},
    package_data={'': ['data/*']},
    include_package_data=True,
    ext_modules=[cjpeglib[v] for v in libjpeg_versions],
    cmdclass={
        "build_ext": custom_build_ext,
        **custom_bdist_wheel
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Information Technology',
        'Intended Audience :: Other Audience',
        'License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: C',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Topic :: Education',
        'Topic :: Multimedia',
        'Topic :: Multimedia :: Graphics',
        'Topic :: Multimedia :: Graphics :: Graphics Conversion',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Image Processing',
        'Topic :: Scientific/Engineering :: Image Recognition',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Security',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Utilities',
    ],
)




import ctypes
from typing import List
from ._bind import CJpegLib


class version:
    """Class grouping functions for controlling libjpeg method."""

    @classmethod
    def set(cls, version: str):
        """Sets the version of libjpeg to use. Loads the library.

        :param version: libjpeg version, one of 6b, 8d, 9d, turbo210.
        :type version: str
        :raises [NotImplementedError]: unsupported libjpeg version

        :Example:

        >>> import jpeglib
        >>> jpeglib.version.set('8d')
        """
        try:
            # libjpeg
            if version in {
                '6b', '7', '8', '8a', '8b', '8c', '8d', '9', '9a', '9b', '9c', '9d', '9e'
            }:
                CJpegLib.set_version(version=version)
            # libjpeg-turbo
            elif version in {
                'turbo120', 'turbo130', 'turbo140', 'turbo150', 'turbo200', 'turbo210'
            }:
                CJpegLib.set_version(version=version)
            # mozjpeg
            elif version in {
                'mozjpeg101', 'mozjpeg201', 'mozjpeg300', 'mozjpeg403'
            }:
                CJpegLib.set_version(version=version)
            else:
                raise NotImplementedError(
                    f'Unsupported libjpeg version: {version}')
        except RuntimeError:
            raise RuntimeError(f'version "{version}" not found, was the package compiled correctly?')

    @staticmethod
    def get() -> str:
        """Gets the currently used version of libjpeg.

        :return: libjpeg version or None if not been loaded yet.
        :rtype: str, None

        :Example:

        >>> import jpeglib
        >>> jpeglib.version.set('6b')
        >>> jpeglib.version.get()
        '6b'
        """
        return CJpegLib.get_version()

    @staticmethod
    def _jpeg_lib_version() -> str:
        """Returns value of jpeg_lib_version macro."""
        return CJpegLib.jpeg_lib_version()

    @staticmethod
    def _get_lib() -> ctypes.CDLL:
        """Low-level getter of the dynamic library.

        :return: Dynamic library object or None if not loaded yet.
        :rtype: ctypes.CDLL, None
        """
        return CJpegLib.get()

    @staticmethod
    def versions() -> List[str]:
        """Lists present DLLs of versions."""
        return CJpegLib.versions()

    def __init__(self, version):
        """Constructor, used in with statement.

        :param version: Version to set inside with block.
        :type version: str
        """
        self.next = version

    def __enter__(self):
        """Sets new version in a block.

        :Example:

        >>> jpeglib.version.set('6b')
        >>> # working with 6b
        >>> # [...]
        >>> with jpeglib.version('8d'):
        >>>     # working with 8d
        >>>     # [...]
        >>>     pass
        """
        self.prev = self.get()
        self.set(self.next)

    def __exit__(self, *args, **kw):
        """Recovers a previous version, when exiting `with` block.

        :Example:

        >>> # working with 6b
        >>> # [...]
        >>> with jpeglib.version('8d'):
        >>>     # working with 8d
        >>>     # [...]
        >>>     pass
        >>> # working with 6b (again)
        >>> # [...]
        """
        self.set(self.prev)


__all__ = ['version']




from .spatial import blockify_8x8, grayscale, luminance
from .dct2D_image import forward_dct, backward_dct
from .dct2D import DCT2D, iDCT2D
__all__ = [
	'blockify_8x8',
	'forward_dct', 'backward_dct',
	'grayscale', 'luminance',
	'DCT2D', 'iDCT2D',
]
